"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import logging
import sqlite3
import traceback
from threading import Thread
from typing import List, Dict
from kudubot.users.Contact import Contact
from kudubot.entities.Message import Message
from kudubot.users.AddressBook import AddressBook
from kudubot.users.Authenticator import Authenticator
from kudubot.users.LanguageSelector import LanguageSelector
from kudubot.exceptions import InvalidConfigException
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


class Connection(object):
    """
    Abstract class that provides a unified model for a chat service connection
    It keeps track of various state variables and provides APIs to send
    messages and start a listener that reacts on incoming messages using the
    implemented service modules
    """

    def __init__(self, services: List[type],
                 config_handler: GlobalConfigHandler):
        """
        Initializes the connection object using the specified services
        Starts the database connection

        :param services: The services to use with the connection
        :param config_handler: The GlobalConfigHandler to determine config
                               file locations
        """
        self.identifier = self.define_identifier()
        self.config_handler = config_handler
        self.logger = logging.getLogger(self.__class__.__module__)

        self.database_file_location = config_handler.data_location
        self.config_file_location = \
            config_handler.specific_connection_config_location
        self.external_services_directory = os.path.join(
            config_handler.external_services_directory, self.identifier
        )

        if not os.path.isdir(self.external_services_directory):
            os.makedirs(self.external_services_directory)

        self.database_file_location = \
            os.path.join(self.database_file_location, self.identifier + ".db")
        self.config_file_location = \
            os.path.join(self.config_file_location, self.identifier + ".conf")

        try:

            self.db = self.get_database_connection_copy()

            self.address_book = AddressBook(self.db)
            self.authenticator = Authenticator(self.db)
            self.language_selector = LanguageSelector(self.db)
            self.config = self.load_config()
            self.user_contact = self.define_user_contact()

            self.services = []
            for service in services:
                try:
                    self.services.append(service(self))
                except Exception as e:
                    # noinspection PyUnresolvedReferences
                    self.logger.error(
                        "Service " + service.define_identifier() +
                        " failed to load due to error:" +
                        str(e.args) + ", traceback:" + traceback.format_exc()
                    )

        except InvalidConfigException as e:
            self.generate_configuration()
            raise e

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the connection's identifier

        :return: The identifier for the Connection type
        """
        raise NotImplementedError()

    def apply_services(self, message: Message, break_on_match: bool = True):
        """
        Applies the services to a Message
        First, the method checks if a service is applicable to a message.
        Then, if it is applicable, the service will process the message
        If the break_on_match parameter is set to True, the first match will
        always end the loop.

        :param message: The message to process
        :param break_on_match: Can be set to False
                               to allow more than one result
        """

        self.logger.info("Received message " + message.message_body + ".")
        self.logger.info("Checking for contact information")

        message.sender = \
            self.address_book.add_or_update_contact(message.sender)
        if message.sender_group is not None:
            message.sender_group = \
                self.address_book.add_or_update_contact(message.sender_group)

        if self.authenticator.is_blacklisted(message.sender) \
                or self.authenticator.is_blacklisted(message.sender_group):
            self.logger.info("Contact is blacklisted. Not applying services")

        self.logger.debug(
            "Applying services to " + repr(message.message_body) + "."
        )

        for service in self.services:
            try:
                if service.is_applicable_to_with_log(message):
                    service.handle_message_with_log(message)
                    if break_on_match:
                        break
            except Exception as e:
                self.logger.error(
                    "Service " + service.identifier +
                    " failed in executing message " + message.message_body +
                    " with exception " + str(e.args) +
                    ", traceback:" + traceback.format_exc())

    def load_config(self) -> Dict[str, object]:
        """
        Loads the configuration for the connection.
        If this fails forsome reason, an InvalidConfigException is raised

        :return: A dictionary containing the configuration
        """
        raise NotImplementedError()

    def generate_configuration(self):
        """
        Generates a new configuration file for this connection.
        """
        raise NotImplementedError()

    def define_user_contact(self) -> Contact:
        """
        Creates a Contact object for the connection by which the
        connection itself is identified

        :return: The connection's user object
        """
        raise NotImplementedError()

    def send_message(self, message: Message):
        """
        Sends a Message using the connection

        :param message: The message to send
        """
        raise NotImplementedError()

    def send_audio_message(self, receiver: Contact, audio_file: str,
                           caption: str = ""):
        """
        Sends an audio message using the connection

        :param receiver: The receiver of the message
        :param audio_file: The path to the audio file to send
        :param caption: The caption sent together with the message
        """
        raise NotImplementedError()

    def send_video_message(self, receiver: Contact, video_file: str,
                           caption: str = ""):
        """
        Sends a video message using the connection

        :param receiver: The recipient of the video message
        :param video_file: The path to the video file to be sent
        :param caption: The caption to be displayed with the video
        """
        raise NotImplementedError()

    def send_image_message(self, receiver: Contact, image_file: str,
                           caption: str = ""):
        """
        Sends an image message using the connection

        :param receiver: The recipient of the image message
        :param image_file: The path to the image file
        :param caption: The caption to be displayed with the image
        """
        raise NotImplementedError()

    def listen(self):
        """
        Starts listening on the connection in an infinite loop.
        If the execution of the program has to continue past starting the
        listener, the listen_in_separate_thread() method should be called
        instead.
        """
        raise NotImplementedError()

    def listen_in_separate_thread(self) -> Thread:
        """
        Runs the listen() method in a separate daemon thread

        :return: The listening thread
        """

        self.logger.info(
            "Starting daemon thread for connection " + self.identifier + "."
        )

        thread = Thread(target=self.listen)
        thread.daemon = True
        thread.start()
        return thread

    def get_database_connection_copy(self) -> sqlite3.Connection:
        """
        Creates a new sqlite Connection to the
        kudubot Connection's database file

        :return: The generated sqlite Connection
        """
        return sqlite3.connect(self.database_file_location)



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import time
import json
import logging
import requests
from typing import List, Dict


logger = logging.getLogger(__name__)
"""
The logger for this module
"""


def scrape_reddit_discussion_threads() -> List[Dict[str, str or int]]:
    """
    Scrapes /u/Holo_of_Yoitsu's post history for anime discussion threads
    :return: The parsed discussion threads as a list of dictionaries with
             descriptive keys
    """

    logger.debug("Fetching last 25 submitted reddit threads "
                 "from /u/Holo_of_Yoitsu")

    html = requests.get("https://www.reddit.com/user/Holo_of_Yoitsu.json")
    while html.status_code != 200:  # Circumvent rate limiting if we're unlucky
        time.sleep(15)
        html = requests.get("https://www.reddit.com/user/Holo_of_Yoitsu.json")

    data = json.loads(html.text)

    threads = []

    for item in data["data"]["children"]:

        url = item["data"]["url"]
        title = item["data"]["title"].split("[Spoilers] ", 1)[1]\
            .rsplit(" discussion", 1)[0]

        show_name = title.rsplit(" - ", 1)[0]
        episode = int(title.rsplit(" - Episode ", 1)[1])

        threads.append({
            "show_name": show_name, "episode": episode, "url": url
        })

        logger.debug("Found thread: " + str(threads[-1]))

    logger.debug("Finished retrieving data from reddit")

    return threads



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Dict

from kudubot.connections.Connection import Connection
from kudubot.entities.Message import Message
from kudubot.users.Contact import Contact


class DummyConnection(Connection):
    """
    A class that implements a Connection for use in unit tests
    """

    @staticmethod
    def define_identifier() -> str:
        return "dummyconnection"

    def listen(self):
        pass

    def define_user_contact(self) -> Contact:
        pass

    def load_config(self) -> Dict[str, object]:
        pass

    def send_message(self, message: Message):
        pass

    def generate_configuration(self):
        pass

    def send_image_message(
            self, receiver: Contact, image_file: str, caption: str = ""):
        pass

    def send_video_message(
            self, receiver: Contact, video_file: str, caption: str = ""):
        pass

    def send_audio_message(
            self, receiver: Contact, audio_file: str, caption: str = ""):
        pass



# -*- coding: utf-8 -*-
"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from typing import List
from threading import Thread
from kudubot.entities.Message import Message


class BaseService(object):
    """
    A class that defines how a chat bot service integrates with a Connection.
    Includes various helper methods.
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the unique identifier for the service

        :return: The Service's identifier.
        """
        raise NotImplementedError()

    @staticmethod
    def define_requirements() -> List[str]:
        """
        Defines the requirements for the service

        :return: The required services for this Service.
        """
        raise NotImplementedError()

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the Service is applicable to a given message

        :param message: The message to check
        :return: True if applicable, else False
        """
        raise NotImplementedError()

    def handle_message(self, message: Message):
        """
        Handles the message, provided this service is applicable to it

        :param message: The message to process
        :return: None
        """
        raise NotImplementedError()

    def __init__(self, connection):
        """
        Initializes the Service using a specified Connection

        :param connection: The connection used by this service
        """
        self.connection = connection
        self.identifier = self.define_identifier()
        self.requires = self.define_requirements()
        self.logger = logging.getLogger(self.__class__.__module__)
        self.init()
        self.logger.debug(self.identifier + " Service initialized")

    # noinspection PyMethodMayBeStatic
    def init(self):
        """
        Helper method that runs after the initialization of the Service object.
        Can be used for anything, but normal use cases would include
        initializing a database table or starting a background thread

        :return: None
        """
        pass

    # -------------------------------------------------------------------------
    #                  _,-'/-'/                                Here be dragons!
    #  .      __,-; ,'( '/
    #   \.    `-.__`-._`:_,-._       _ , . ``
    #    `:-._,------' ` _,`--` -: `_ , ` ,' :
    #       `---..__,,--'            ` -'. -'
    # Everything below this should not be overridden by subclasses
    # -------------------------------------------------------------------------

    def reply(self, title: str, body: str, message: Message):
        """
        Provides a helper method that streamlines the process
        of replying to a message.
        Very useful for Services that send a reply immediately
        to cut down on clutter in the code

        :param title: The title of the message to send
        :param body: The body of the message to send
        :param message: The message to reply to
        :return: None
        """
        reply_message = Message(
            title, body, message.get_direct_response_contact(),
            self.connection.user_contact
        )
        self.connection.send_message(reply_message)

    def is_applicable_to_with_log(self, message: Message) -> bool:
        """
        Wrapper around the is_applicable_to method,
        which enables logging the message easily
        for all subclasses

        :param message: The message to analyze
        :return: True if the message is applicable, False otherwise
        """
        self.logger.debug("Checking if " + message.message_body +
                          " is applicable")

        # Check if a subclass is_applicable fits here
        # This construct is required to avoid having to call super()
        # in implemented service classes, which can not be enforced or
        # hinted at via an IDE
        for subclass in [
            "multi_language", "authenticated", "helper", "external"
        ]:
            is_applicable_method = \
                getattr(self, "is_applicable_to_" + subclass, None)
            if callable(is_applicable_method):
                if is_applicable_method(message):
                    self.logger.debug("Message is applicable")
                    return True
                elif subclass == "authenticated":
                    self.reply("Access Denied", "Access Denied ❌", message)
                    return False

        result = self.is_applicable_to(message)
        not_word = "" if result else "not"
        self.logger.debug("Message is " + not_word + " applicable")
        return result

    def handle_message_with_log(self, message: Message):
        """
        Wrapper around the handle_message method,
        which enables logging the message easily
        for all subclasses

        :param message: The message to handle
        :return: None
        """
        self.logger.debug("Handling message " + message.message_body)

        # Checks if a parent class of the service is applicable, if yes,
        # Runs their handle_message method
        for subclass in [
            "multi_language", "authenticated", "helper", "external"
        ]:
            is_applicable_method = \
                getattr(self, "is_applicable_to_" + subclass, None)
            handle_method = getattr(self, "handle_message_" + subclass, None)

            if callable(is_applicable_method) and callable(handle_method):

                if is_applicable_method(message):
                    handle_method(message)
                    return

        self.handle_message(message)

    # noinspection PyMethodMayBeStatic
    def start_daemon_thread(self, target: callable) -> Thread:
        """
        Starts a daemon/background thread
        :param target: The target function to execute as a separate thread
        :return: The thread
        """
        self.logger.debug("Starting background thread for " + self.identifier)

        thread = Thread(target=target)
        thread.daemon = True
        thread.start()
        return thread

    # noinspection PyDefaultArgument
    def initialize_database_table(self, sql: List[str] = [],
                                  initializer: callable=None):
        """
        Executes the provided SQL queries to create the database table(s).

        :param sql: The SQL queries used to create the database tables
        :param initializer: A method that initializes
                            the database connection itself.
        :return: None
        """
        self.logger.debug("Initializing Database table" +
                          ("" if len(sql) < 2 else "s") +
                          " for " + self.identifier)
        for statement in sql:
            self.connection.db.execute(statement)
        if initializer is not None:
            initializer(self.connection.db)



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import stat
import json
from typing import List
from subprocess import Popen


def build(service_directory):
    """
    Builds a Service using the service configuration
    inside the service directory.

    :param service_directory: The location of the service directory
    :return: The path to the generated executable file
    """

    service_name = os.path.basename(service_directory)
    current_dir = os.getcwd()
    os.chdir(service_directory)

    with open("service.json", 'r') as f:
        config = json.load(f)

    try:
        Popen(config["build_commands"]).wait()
    except BaseException as e:
        print(e)
        os.chdir(current_dir)
        return

    output = config["output_file"]
    st = os.stat(output)
    os.chmod(output, st.st_mode | stat.S_IEXEC)  # Make executable

    if os.path.basename(output) != service_name:
        new_output = os.path.join(os.path.dirname(output), service_name)
        os.rename(output, new_output)
        output = new_output

    os.chdir(current_dir)

    return os.path.join(service_directory, output)


def build_external(move_to: str = "") -> List[str]:
    """
    Builds all external services

    :param move_to: If specified, all built files are moved to that directory
    :return: A list of all generated executable files
    """

    external_dir = os.path.join("kudubot", "services", "external")
    built_executables = []

    for service in os.listdir(external_dir):

        service_dir = os.path.join(external_dir, service)
        if not os.path.isdir(service_dir) or service == "__pycache__":
            continue

        result = build(service_dir)

        if move_to != "" and os.path.isfile(result):
            destination = os.path.join(move_to, service)
            os.rename(result, destination)
            built_executables.append(destination)

        elif os.path.isfile(result):
            built_executables.append(result)

    return built_executables



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import shutil
from kudubot.config.GlobalConfigHandler import GlobalConfigHandler


def generate_test_environment() -> GlobalConfigHandler:
    """
    Generates the test environment

    :return: The GlobalConfigHandler that points to the configuration
    """

    handler = GlobalConfigHandler("kudu-test")
    handler.generate_configuration()
    return handler


def clean_up_test_environment():
    """
    Deletes the test environment

    :return: None
    """
    if os.path.isdir("kudu-test"):
        shutil.rmtree("kudu-test")



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import sqlite3
from typing import List, Dict


logger = logging.getLogger(__name__)
"""
The logger for this module
"""


# noinspection SqlNoDataSourceInspection,SqlDialectInspection
def initialize_database(db: sqlite3.Connection):
    """
    Initializes the Service's database tables

    :return: None
    """
    logger.info("Initializing Anime Reminder Database")

    db.execute("CREATE TABLE IF NOT EXISTS anime_reminder_subscriptions ("
               "    id INTEGER CONSTRAINT constraint_name PRIMARY KEY,"
               "    user_id INTEGER NOT NULL,"
               "    show_name VARCHAR(255) NOT NULL"
               ")")
    db.execute("CREATE TABLE IF NOT EXISTS anime_reminder_threads ("
               "    id INTEGER CONSTRAINT constraint_name PRIMARY KEY,"
               "    show_name VARCHAR(255) NOT NULL,"
               "    episode INTEGER NOT NULL,"
               "    thread VARCHAR(255) NOT NULL"
               ")")
    db.commit()


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def subscription_exists(user: int, show_name: str, db: sqlite3.Connection) \
        -> bool:
    """
    Checks if a subscription already exists

    :param user: The user ID for this subscription
    :param show_name: The show name of the subscription
    :param db: The database to use
    :return: True if the subscription already exists, otherwise False
    """
    subscriptions = filter(
        lambda x: x["show_name"].lower() == show_name.lower(),
        get_subscriptions(db, user)
    )
    return len(list(subscriptions)) >= 1


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def store_subscription(user: int, show_name: str, db: sqlite3.Connection) \
        -> bool:
    """
    Creates a new subscription if it does not exist yet

    :param user: The user ID for this subscription
    :param show_name: The show name of the subscription
    :param db: The database to use
    :return: True if a new subscription was stored,
             False if the subscription already existed
    """
    exists = subscription_exists(user, show_name, db)

    if not exists:
        logger.debug("Creating new subscription for user " + str(user) +
                     " and show " + show_name)
        db.execute("INSERT INTO anime_reminder_subscriptions "
                   "(user_id, show_name) VALUES (?, ?)", (user, show_name))
        db.commit()
    else:
        logger.debug("Subscription for user " + str(user) + " and show " +
                     show_name + " already exists")

    return not exists


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def delete_subscription(user: int, show_name: str, db: sqlite3.Connection) \
        -> bool:
    """
    Deletes a subscription from the database

    :param user: The user ID for this subscription
    :param show_name: The show name of the subscription
    :param db: The database to use
    :return True if the subscription existed and is deleted,
            False if the subscription did not exist beforehand
    """
    exists = subscription_exists(user, show_name, db)

    if exists:
        logger.debug("Deleting subscription for user " + str(user) +
                     " and show " + show_name)
        db.execute("DELETE FROM anime_reminder_subscriptions "
                   "WHERE user_id=? AND show_name LIKE ?",
                   (user, show_name))
        db.commit()
    else:
        logger.debug("Subscription for user " + str(user) + " and show " +
                     show_name + " did not exist. Not deleting anything.")

    return exists


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def get_subscriptions(db: sqlite3.Connection, user: int=-1) \
        -> List[Dict[str, str or int]]:
    """
    Fetches a list of subscriptions from the database, either for all users or
    one specific user.

    :param db: The database to use
    :param user: The user parameter can limit the query to search for
                 subscriptions for only one user
    :return: A list of dictionaries representing the subscriptions
    """

    if user == -1:
        subs = db.execute("SELECT id, user_id, show_name "
                          "FROM anime_reminder_subscriptions").fetchall()

    else:
        subs = db.execute("SELECT id, user_id, show_name "
                          "FROM anime_reminder_subscriptions "
                          "WHERE user_id=?", (user,)).fetchall()

    formatted_subs = []
    for sub in subs:
        formatted_subs.append({
            "id": sub[0], "user_id": sub[1], "show_name": sub[2]
        })

    return formatted_subs


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def thread_exists(thread: Dict[str, str or int], db: sqlite3.Connection) \
        -> bool:
    """
    Checks if a reddit discussion thread with the given thread parameters
    already exists

    :param thread: The thread to check for
    :param db: The database to check
    :return: True if the thread exists, false otherwise
    """
    result = db.execute(
        "SELECT * FROM anime_reminder_threads "
        "WHERE show_name=? AND episode=? AND thread=?",
        (thread["show_name"], thread["episode"], thread["url"])
    )
    return len(result.fetchall()) > 0


# noinspection SqlNoDataSourceInspection,SqlDialectInspection,SqlResolve
def store_thread(thread: Dict[str, str or int], db: sqlite3.Connection):
    """
    Stores a reddit discussion thread in the database

    :param thread: The thread to store
    :param db: The database to use
    :return: None
    """
    logger.debug("Storing reddit thread " + str(thread))
    db.execute(
        "INSERT INTO anime_reminder_threads (show_name, episode, thread) "
        "VALUES (?, ?, ?)",
        (thread["show_name"], thread["episode"], thread["url"])
    )
    db.commit()



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import List, Dict
from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService


class ServiceListerService(HelperService):
    """
    This Service lists all currently active services
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the identifier for this service

        :return: The service's identifier
        """
        return "service_lister"

    def define_command_name(self, language: str) -> str:
        """
        Defines the command prefix for the service.
        By default, '/' followed by the service identifier from
        self.define_identifier() is used.

        :param language: The language in which to define the command name
        :return: The command name in the specified language
        """
        return {
            "en": "/services",
            "de": "/dienste"
        }[language]

    @staticmethod
    def define_requirements() -> List[str]:
        """
        Defines required Services for this Service

        :return: The list of required services
        """
        return []

    def determine_language(self, message: Message) -> str:
        """
        Determines the language of a message

        :param message: The message to check for the language
        :return: The language of the message
        """
        try:
            return {
                "/services": "en", "/dienste": "de"
            }[message.message_body.strip().split(" ")[0]]
        except KeyError:
            return "en"

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the dictionary used for translating strings using the
        self.reply_translated() or self.translate() methods

        The format is:

            term: {lang: value}

        :return: The dictionary for use in translating
        """
        return {
            "@{Title}": {"en": "Service List", "de": "Dienstliste"},
            "@{BodyTitle}": {"en": "Active Services", "de": "Aktive Dienste"}
        }

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message of the service in various languages

        :param language: The language to use
        :return: The help message in the language
        """
        return {
            "en": "Use the `/services` command to get a list of active "
                  "services on this kudubot instance",
            "de": "Benutze den `/dienste` Befehl um eine Liste aktiver "
                  "Dienste auf dieser kudubot Instanz zugesendet zu bekommen."
        }[language]

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the Syntax description of the Service in various languages

        :param language: The language in which to return the syntax message
        :return: The syntax message
        """
        return {
            "en": "/services",
            "de": "/dienste"
        }[language]

    def handle_message(self, message: Message):
        """
        Handles an applicable message.
        Sends a message containing the identifiers of all active services

        :param message: The message to handle
        """
        reply_msg = "@{BodyTitle}:\n\n"
        for service in self.connection.services:
            reply_msg += service.identifier + "\n"
        self.reply_translated("@{TITLE}", reply_msg.strip(), message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if a Message is applicable to this Service
        Checks if the

        :param message: The message to check
        :return: True if the Message is applicable, False otherwise
        """
        return message.message_body.strip() in [
            "/dienste", "/services"
        ]



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

from typing import Dict


class Contact(object):
    """
    Class that models a contact in the connection's address book database
    """

    def __init__(self, database_id: int, display_name: str, address: str):
        """
        Initializes the contact object

        :param database_id: The contact's ID in the database
        :param display_name: The display name of the contact
        :param address: The contact's address
        """
        self.database_id = database_id
        self.display_name = display_name
        self.address = address

    def to_dict(self) -> Dict[str, str or int]:
        """
        :return: The contact as a dictionary,
                 which can be used to store the contact in a JSON file
        """
        return {
            "database_id": self.database_id,
            "display_name": self.display_name,
            "address": self.address
        }


def from_dict(data: Dict[str, str or int]) -> Contact:
    """
    Generates a Contact object from a dictionary

    :param data: The data to turn into a Contact object
    :return: The generated Contact object
    """
    return Contact(
        data["database_id"],
        data["display_name"],
        data["address"]
    )



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""


import re
import time
import requests
from typing import List, Dict
from kudubot.entities.Message import Message
from kudubot.services.HelperService import HelperService
from kudubot.services.native.anime_reminder.scraper import \
    scrape_reddit_discussion_threads
from kudubot.services.native.anime_reminder.database import \
    initialize_database, store_subscription, delete_subscription, \
    get_subscriptions, thread_exists, store_thread


class AnimeReminderService(HelperService):
    """
    The Kudubot Service that provides the anime reminder functionality
    """

    def init(self):
        """
        In addition to the normal initialization of a Service,
        this service initializes
        its database and starts the background thread
        """
        self.initialize_database_table(initializer=initialize_database)
        self.start_daemon_thread(self.background_loop)

    @staticmethod
    def define_requirements() -> List[str]:
        """
        :return: The dependencies for this service
        """
        return []

    @staticmethod
    def define_identifier() -> str:
        """
        :return: The service's identifier
        """
        return "anime_reminder"

    def handle_message(self, message: Message):
        """
        Handles a message from the user

        :param message: The message to handle
        :return: None
        """
        mode = message.message_body.split(" ")[1].lower()
        user = message.get_direct_response_contact()
        user_id = user.database_id

        params = message.message_body.split("\"")
        show = "" if len(params) < 2 else params[1]

        if mode in [self.translate("@subscribe_command", x)
                    for x in self.supported_languages()]:

            result = "@successful_add_message" \
                if store_subscription(user_id, show, self.connection.db) \
                else "@exists_message"
            self.reply_translated("@reply_title", "@subscribe_message_start " +
                                  show + " " + result + ".", message)

        elif mode in [self.translate("@unsubscribe_command", x)
                      for x in self.supported_languages()]:

            result = "@successful_remove_message" \
                if delete_subscription(user_id, show, self.connection.db) \
                else "@does_not_exist_message"
            self.reply_translated("@reply_title",
                                  "@unsubscribe_message_start " + show +
                                  " " + result + ".", message)

        elif mode in [self.translate("@list_command", x)
                      for x in self.supported_languages()]:

            message_text = "@subscription_list_message:\n\n"
            for subscription in get_subscriptions(self.connection.db, user_id):
                message_text += subscription["show_name"] + "\n"
            self.reply_translated("@reply_title", message_text, message)

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if a message is applicable to this Service

        :param message: The message to check
        :return: None
        """
        applicable = False
        for language in self.supported_languages():

            regex = "^@command_name " \
                    "(@list_command|@sub_unsub_command \"[^\"]+\")$"
            regex = re.compile(self.translate(regex, language))
            applicable = bool(re.search(regex, message.message_body)) \
                or applicable

        return applicable

    def define_syntax_description(self, language: str) -> str:
        """
        Defines the syntax description for the Service

        :param language: The language in which to display the syntax
                         description
        :return: The syntax description in the specified language
        """
        skeleton = "@command_name @list_command\n"
        skeleton += "@command_name @subscribe_command " \
                    "\"@show_name_parameter\"\n"
        skeleton += "@command_name @unsubscribe_command " \
                    "\"@show_name_parameter\""
        return self.translate(skeleton, language)

    def define_language_text(self) -> Dict[str, Dict[str, str]]:
        """
        Defines the Language strings for the service for the implemented
        languages

        :return: The dictionary used to translate strings
        """
        return {
            "@list_command": {"en": "list",
                              "de": "auflisten"},
            "@subscribe_command": {"en": "subscribe",
                                   "de": "abonnieren"},
            "@unsubscribe_command": {"en": "unsubscribe",
                                     "de": "abbestellen"},
            "@sub_unsub_command": {"en": "(un)?subscribe",
                                   "de": "(abonnieren|abbestellen)"},
            "@show_name_parameter": {"en": "Show Name",
                                     "de": "Serienname"},
            "@command_name": {"en": "/anime-remind",
                              "de": "/anime-erinner"},
            "@exists_message": {"en": "already exists",
                                "de": "existiert bereits"},
            '@does_not_exist_message': {"en": "does not exist",
                                        "de": "existiert nicht"},
            '@successful_add_message': {"en": "successful",
                                        "de": "wurde erfolgreich abonniert"},
            '@successful_remove_message': {"en": "successfully removed",
                                           "de": "wurde erfolgreich "
                                                 "abbestellt"},
            "@subscribe_message_start": {"en": "Subscription for",
                                         "de": "Abonnement für"},
            "@unsubscribe_message_start": {"en": "Subscription for",
                                           "de": "Abonnement für"},
            "@reply_title": {"en": "Anime Reminder",
                             "de": "Anime Erinnerung"},
            "@remind_title": {"en": "Anime Reminder",
                              "de": "Anime Erinnerung"},
            "@subscription_list_message": {"en": "You are subscribed to the"
                                                 "following shows",
                                           "de": "Du hast folgende Serien"
                                                 "abonniert"},
            "@episode": {"en": "episode",
                         "de": "Episode"},
            "@has_been_released": {"en": "has been released! Discuss it"
                                         "on reddit",
                                   "de": "wurde veröffentlicht! Diskuttier"
                                         "sie auf reddit"},
            "@help_message": {
                "en": "The Anime Reminder Service periodically checks for new "
                      "anime discussion threads on reddit's "
                      "r/anime board. A user can subscribe to specific shows "
                      "and will then be notified whenever a "
                      "new thread appears, which usually coincides with the "
                      "point in time that the episode is "
                      "available on streaming services like Crunchyroll, "
                      "Amazon or Funimation.\n"
                      "Users can also unsubscribe from shows and list all "
                      "their subscriptions. Subscription names "
                      "are case-insensitive.",
                "de": "Der Anime Erinnerungs-Service schaut in periodischen "
                      "Abständen nach, ob neue Anime Diskussionen "
                      "auf reddit.com/r/anime entstanden sind, welche neue "
                      "Anime Episode diskuttieren. Ein Nutzer kann "
                      "einzelne Serien abonnieren und wieder abbestellen. "
                      "Für abonnierte Serien erhält der Nutzer "
                      "eine Nachricht sobal eine neue Episode verfügbar ist."
                      "Die Namen für die Abonnements achten nicht auf Groß- "
                      "und Kleinschreibung. Man kann auch alle"
                      "Abonnements die derzeit aktiv sind auflisten lassen."
            }
        }

    def define_help_message(self, language: str) -> str:
        """
        Defines the help message for the service

        :param language: The target language to translate the message to
        :return: The help message in the specified language
        """
        return self.translate("@help_message", language)

    def determine_language(self, message: Message) -> str:
        """
        Determines the language used based on the incoming message

        :param message: The message to analyze
        :return: The language key to use
        """
        if message.message_body.startswith(self.define_command_name("de")):
            return "de"
        else:
            return "en"

    def define_command_name(self, language: str) -> str:
        """
        Defines the command name for the anime-remind functionality

        :param language: The language for which the command is valid
        :return: The command name in the specified language
        """
        return self.translate("@command_name", language)

    def background_loop(self):
        """
        Starts a new thread which will continuously check for new anime
        discussion threads on reddit.com/r/anime
        and notify users if they are subscribed to those shows

        :return: None
        """
        db = self.connection.get_database_connection_copy()
        while True:

            try:
                new_threads = scrape_reddit_discussion_threads()
            except requests.exceptions.ConnectionError:
                self.logger.error("Failed to connect")
                time.sleep(60)
                continue

            self.logger.debug("Checking for due subscriptions")

            for thread in new_threads:

                if not thread_exists(thread, db):
                    store_thread(thread, db)

                    for subscription in get_subscriptions(db):
                        if subscription["show_name"].lower() \
                                == thread["show_name"].lower():

                            receiver = self.connection.address_book\
                                .get_contact_for_id(
                                    subscription["user_id"], db
                                )

                            message_text = thread["show_name"] + " @episode "
                            message_text += str(thread["episode"])
                            message_text += " @has_been_released: "
                            message_text += thread["url"]

                            language = self.connection.language_selector.\
                                get_language_preference(
                                    receiver.database_id, db=db
                                )

                            message_text = \
                                self.translate(message_text, language)
                            title = self.translate("@remind_title", language)

                            self.connection.send_message(Message(
                                title, message_text, receiver,
                                self.connection.user_contact)
                            )

            time.sleep(60)



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from typing import List
from kudubot.entities.Message import Message
from kudubot.services.BaseService import BaseService


class SimpleResponderService(BaseService):
    """
    Class that implements a kudubot service that analyzes Strings and responds
    to them using a relatively simple ruleset
    """

    rules = [
        # Case Insensitive Equals
        ({":)": ":) :) :)"},
         lambda incoming, key: incoming.lower() == key.lower()),

        # Case Sensitive Equals
        ({"Hello World!": "01001000 01100101 01101100 01101100 01101111 "
                          "00100000 01010111 01101111 01110010 01101100 "
                          "01100100 00100001",
          "Ping": "Pong",
          "ping": "pong"},
         lambda incoming, key: incoming == key),

        # Case Insensitive Contains
        ({"würfel": "4"},
         lambda incoming, key: incoming.lower().find(key.lower()) != -1),

        # Case Sensitive Contains
        ({"FC Bayern": "Deutscher Meister 2017!"},
         lambda incoming, key: incoming.find(key) != -1)
    ]
    """
    These are the rule-based responses for the Simple Responder Service

    The consist of a Tuple of a dictionary mapping strings to each other,
    as well as a lambda expression that return a boolean value.
    If that lambda expression returns true, the pattern matches and the
    dictionary value is returned as a response
    """

    @staticmethod
    def define_identifier() -> str:
        """
        Defines the Service's unique identifier

        :return: The unique identifier
        """
        return "simple_responder"

    @staticmethod
    def define_requirements() -> List[str]:
        """
        Defines the Service's dependency Services

        :return: A list of dependency Services for this Service
        """
        return []

    def __check_rules__(self, message_text) -> str:
        """
        Goes through all rules and returns the first appropriate response text

        :param message_text: The message text to analyze
        :return: The response, or an empty string if nothing was found
        """

        self.logger.debug("Checking Rules")

        for rule in self.rules:
            for sub_rule in dict(rule[0]):
                if rule[1](message_text, sub_rule):
                    logging.debug("Match found for rule " + repr(sub_rule))
                    return rule[0][sub_rule]

        return ""

    def is_applicable_to(self, message: Message) -> bool:
        """
        Checks if the Service is applicable to a given message

        :param message: The message to check
        :return: True if applicable, else False
        """
        return self.__check_rules__(message.message_body) != ""

    def handle_message(self, message: Message):
        """
        Handles the message, provided this service is applicable to it

        :param message: The message to process
        :return: None
        """
        response = self.__check_rules__(message.message_body)
        if response != "":
            self.reply("Simple Response", response, message)



"""
Copyright 2015-2017 Hermann Krumrey

This file is part of kudubot.

kudubot is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

kudubot is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with kudubot.  If not, see <http://www.gnu.org/licenses/>.
"""

import sqlite3
import logging
from typing import Dict, List
from datetime import datetime
from kudubot.users.Contact import Contact


logger = logging.getLogger(__name__)
"""
The logger for this module
"""


def initialize_database(database: sqlite3.Connection):
    """
    Initializes the Database Table for the reminder service

    :param database: The database connection to use
    :return: None
    """
    # noinspection SqlNoDataSourceInspection,SqlDialectInspection
    database.execute("CREATE TABLE IF NOT EXISTS reminder ("
                     "    id INTEGER CONSTRAINT constraint_name PRIMARY KEY,"
                     "    sender_id INTEGER NOT NULL,"
                     "    msg_text VARCHAR(255) NOT NULL,"
                     "    due_time VARCHAR(255) NOT NULL,"
                     "    sent BOOLEAN NOT NULL"
                     ")")
    database.commit()


def convert_datetime_to_string(to_convert: datetime) -> str:
    """
    Converts a datetime object to a string that can be stored in the database

    :param to_convert: The datetime object to convert
    :return: The datetime as a string
    """
    return to_convert.strftime("%Y-%m-%d:%H-%M-%S")


def convert_string_to_datetime(to_convert: str) -> datetime:
    """
    Converts a string of the form '%Y-%m-%d:%H-%M-%S' to a datetime object

    :param to_convert: The string to convert
    :return: the resulting datetime object
    """
    return datetime.strptime(to_convert, "%Y-%m-%d:%H-%M-%S")


# noinspection SqlNoDataSourceInspection,SqlResolve
def get_next_id(database: sqlite3.Connection):
    """
    Fetches the next Reminder ID

    :param database: The database to use
    :return: The next highest reminder ID
    """
    return database.execute(
        "SELECT CASE WHEN COUNT(id) > 0 THEN MAX(id) ELSE 0 END AS max_id "
        "FROM reminder").fetchall()[0][0] + 1


# noinspection SqlNoDataSourceInspection,SqlResolve
def store_reminder(database: sqlite3.Connection, message: str,
                   due_time: datetime, sender_id: int):
    """
    Stores a reminder in the database

    :param database: The database Connection to use
    :param message: The message text to store
    :param due_time: The time at which the message should be sent
    :param sender_id: The initiator's id in the address book table
    :return: None
    """
    database.execute(
        "INSERT INTO reminder (id, sender_id, msg_text, due_time, sent) "
        "VALUES (?, ?, ?, ?, ?)",
        (get_next_id(database), sender_id, message,
         convert_datetime_to_string(due_time), False)
    )
    database.commit()
    logger.info("Reminder stored")


# noinspection SqlNoDataSourceInspection,SqlResolve
def get_unsent_reminders(database: sqlite3.Connection) \
        -> List[Dict[str, str or int or datetime]]:
    """
    Retrieves all unsent reminders from the database

    :param database: The database to use
    :return: A list of dictionaries that contain the reminder information
    """

    results = database.execute(
        "SELECT reminder.id, reminder.msg_text, reminder.due_time, "
        "address_book.address, address_book.id, address_book.display_name "
        "FROM reminder "
        "JOIN address_book ON reminder.sender_id = address_book.id "
        "WHERE reminder.sent = 0")
    formatted_results = []
    for result in results:
        formatted_results.append({
            "id": result[0],
            "message": result[1],
            "due_time": convert_string_to_datetime(result[2]),
            "receiver": Contact(result[4], result[5], result[3])
        })
    return formatted_results


def mark_reminder_sent(database: sqlite3.Connection, reminder_id: int):
    """
    Marks a reminder as sent

    :param database: The database connection to use
    :param reminder_id: The Reminder ID
    :return: None
    """
    logger.info("Marking reminder '" + str(reminder_id) + "' as sent.")
    # noinspection SqlNoDataSourceInspection,SqlResolve
    database.execute("UPDATE reminder SET sent=? WHERE id=?",
                     (True, reminder_id))
    database.commit()
